#!/bin/bash
#
# ACME PKI (Certificate) Bootstrap v1.3
# 
# This script will generate all the files necessary to build a certificate chain of trust
# using a CA, CA Intermediate, Server, and Client certificates. After the bootstrap the other
# helper scripts will generate new client/server certificates
# 

# source this file to include the functions
. res/libs/pki_funcs.sh

PARAM1=$1
PARAM2=$2

usage() {
  echo
  echo "This application will generate all the files necessary to build a certificate chain of trust"
  echo "using a CA, CA Intermediate, Server, and Client certificates. All the files are put into a"
  echo "PKI Lifecycle package. A .cnf file is required for the domain. The domain url should match"
  echo "the .cnf file name. Put the .cnf config file into the .res/cnf/ directory"
  echo 
  echo "Usage: pki_bootstrap  <.cnf file (minus the .cnf)>  [# of CA-I to generate]"
  echo
  echo "Example: pki_bootstrap  org.acme.xyz"
  echo "         pki_bootstrap  org.acme.xyz  5"
  echo
  exit 1
}

# 
# Grab the latest serial # from the file, auto-increment
# 
get_serial_ca() {
  SERIAL=`head res/cfg/SERIAL`
  if [[ -z $SERIAL ]]; then
    SERIAL=11111
    echo_block "WARN: no file 'SERIAL' found, using default 11111 as the serial # for CA"
  else
    PLUS1=$((SERIAL+1))
    echo $PLUS1 > res/cfg/SERIAL
  fi
}

#
# CA generation requires .cnf files
#
app_init() {
  if [[ -n $PARAM1 ]]; then
    # need to know the location of the configuration file  (expected to be in same dir path as this script)
    CA_CNF="$CD_ROOT/res/cnf/ca.cnf"

    # handle the case of having the ".cnf" extension or not
    if [[ ${PARAM1: -4} == .cnf ]]; then
      ORG_URL=${PARAM1%.*}
      S_CNF=${PARAM1}
    else
      ORG_URL=$PARAM1
      S_CNF="${PARAM1}.cnf"
    fi

    FQ_S_CNF="${CD_ROOT}/res/cnf/${S_CNF}"
    if [[ ! -f $FQ_S_CNF ]] || [[ ! -f $CA_CNF ]]; then
      usage
    fi
  else
    usage
  fi
}

#
# Generate a new Certificate Authority 
# Create a new LifeCycle package
# 
# IN: UNIQ_ID_CA, SERIAL
#
mk_lifecycle_pkg() {
  get_serial_ca

  # Organize
  # 
  # create a unique path for the server certificate
  UNIQ_DIR_LC=`date +%Y-%m-%d.%H_%M_%S`
  UNIQ_DIR_LC="pki-lifecycle_${UNIQ_DIR_LC}"
  FQ_DIR_LC=`pwd`
  FQ_DIR_LC="${FQ_DIR_LC}/${UNIQ_DIR_LC}"

  # create CA unique dir
  UNIQ_ID_CA="${SERIAL}.ca.${ORG_URL}"
  mkdir -p "${UNIQ_DIR_LC}/ca"
  cd "${UNIQ_DIR_LC}"

  # generate a new CA
  gen_ca $UNIQ_ID_CA $SERIAL

  # go back to original dir
  cd ..
}

#
#
#
cp_lifecycle_docs() {
  # resource files to be copied to the PKI Lifecycle Package
  RES="${CD_ROOT}/res"
  mkdir -p "${UNIQ_DIR_LC}/cfg"
  echo $UNIQ_ID_CA >         $CD_ROOT/$UNIQ_DIR_LC/cfg/UNIQ_ID_CA
  cp -r $CD_ROOT/res         $CD_ROOT/$UNIQ_DIR_LC/
  cp $RES/libs/gen_ca-i.sh   $CD_ROOT/$UNIQ_DIR_LC/
  cp $RES/docs/README_LC     $CD_ROOT/$UNIQ_DIR_LC/README
  cp $RES/docs/SERIAL_LC     $CD_ROOT/$UNIQ_DIR_LC/cfg/SERIAL
  cp $RES/libs/pki_funcs.sh  $CD_ROOT/$UNIQ_DIR_LC/cfg/
  cp $RES/cnf/$ORG_URL.cnf   $CD_ROOT/$UNIQ_DIR_LC/cfg/
  cp $RES/cnf/ca.cnf         $CD_ROOT/$UNIQ_DIR_LC/cfg/

  # CA certs
  cp $CD_ROOT/$UNIQ_DIR_LC/ca/*.crt.pem   $CD_ROOT/$UNIQ_DIR_LC/cfg/ca.crt.pem
  cp $CD_ROOT/$UNIQ_DIR_LC/ca/*.keys.pem  $CD_ROOT/$UNIQ_DIR_LC/cfg/ca.keys.pem
}

#
# Generate Lifecycle CA Intermediates
#
gen_lc_cai() {
  cd $FQ_DIR_LC

  if [[ -n $PARAM2 ]]; then
    COUNT=$(($PARAM2-1))
  else
    COUNT=1
  fi

  for NUM in $(seq 0 $COUNT)
  do
    ca-i_gen_pki $ORG_URL 5
  done
}

# ***** ***** ***** ***** *****
# 
# CERTIFICATE AUTHORITY (CA)
# 
# ***** ***** ***** ***** *****
# This function will generate a CA Intermediate
# IN: UNIQ_ID_CA, SERIAL
#
gen_ca() {
  UNIQ_ID_CA=$1
  SERIAL=$2

  echo_block "Create CA (${UNIQ_ID_CA})"

  # encrypt the key
  #openssl genrsa -aes256 -password "pass:password" -out ca.keys.pem 4096 

  # key un-protected
  openssl genrsa -out "ca/${UNIQ_ID_CA}.keys.pem" 4096
  # 
  # Create Certificate   (valid for 10 years, after the entire chain of trust expires)
  openssl req -config $CD_ROOT/res/cnf/ca.cnf -new -x509 -sha256 -days 3650 -extensions v3_ca \
              -subj "/C=OO/O=ACME/CN=${UNIQ_ID_CA}" -set_serial ${SERIAL} \
              -key ca/${UNIQ_ID_CA}.keys.pem -out ca/${UNIQ_ID_CA}.crt.pem

  # verify certificate (output to text file for review)
  openssl x509 -noout -text -in ca/${UNIQ_ID_CA}.crt.pem > ca/${UNIQ_ID_CA}_cert.info.txt
}


main() {
  CD_ROOT=`pwd`

  # generate new CA
  # create new PKI Lifecycle Package
  app_init
  mk_lifecycle_pkg
  cp_lifecycle_docs

  # gen some CAs
  gen_lc_cai  
  
  # make sure we return to root execution path
  cd "${CD_ROOT}"
}


# ***** ***** ***** ***** ***** ***** ***** ***** ***** ***** ***** *****
# 
# main execution begins here  (because all the functions have to be defined)
# 
# ***** ***** ***** ***** ***** ***** ***** ***** ***** ***** ***** *****

main

# ***** ***** ***** *****
# 
#
# 
# ***** ***** ***** *****
