#!/bin/bash
#
# all main functions to generate a PKI certificate chain
# 

#
# print text wrapped in a block
#
echo_block() {
  echo
  echo "***** ***** ***** *****"
  echo $1
  echo "***** ***** ***** *****"
}

# 
# Grab the latest serial # from the file, auto-increment
# 
get_serial() {
  SERIAL=`head cfg/SERIAL`
  if [[ -z $SERIAL ]]; then
    SERIAL=11111
    echo_block "WARN: no file 'SERIAL' found, using default 11111 as the serial # for CA"
  else
    PLUS1=$((SERIAL+1))
    echo $PLUS1 > cfg/SERIAL 
  fi
}

#
# check the integrity of the CA-I package
# 
check_cai_pkg() {
  if [[ ! -f cfg/ca.keys.pem ]] || [[ ! -f cfg/ca.crt.pem ]]; then
    if [[ ! -f cfg/ca-i.keys.pem ]] || [[ ! -f cfg/ca-i.crt.pem ]]; then
      echo_block "ERROR: missing a config file: cfg/ca.crt.pem, cfg/ca.keys.pem, cfg/ca-i.crt.pem, cfg/ca-i.keys.pem"
      usage
    fi
  fi
  if [[ ! -f cfg/SERIAL ]]; then
    echo_block "ERROR: file cfg/SERIAL is missing"
    usage
  fi
}

#
# Create CA Intermediate PKI
# 
# 

#
# Generate a PKI chain
#   - the certificate chain is unique based on the serial #
#   - generate a new CA I
#   - generate server certificates
#   - generate client certificates
#
# INPUT:  ORG URL, SERIAL #, LOOP NUM
# 
ca-i_gen_pki() {
  CDD=`pwd`
  ORG_URL=$1
  NUM_CERTS=$2

  # create unique directory
  get_serial
  UNIQ_ID_CAI="${SERIAL}.cai.${ORG_URL}"
  mkdir -p "distribution/${UNIQ_ID_CAI}"

  # generate CA Intermediate
  ca-i_gen_cert $UNIQ_ID_CAI

  # create directories, copy files, before generating client/server
  __ca-i_create_pkg

  # the client & server applications need to execute in their perspective directories
  cd $CDD/distribution/$UNIQ_ID_CAI/clients
  gen_client $NUM_CERTS

  cd $CDD/distribution/$UNIQ_ID_CAI/servers
  gen_server $NUM_CERTS

  # return to last path
  cd $CDD
}

#
# Copies all applcations to the Lifecycle package
#   organize the ca-i directory
#   order matters:  move these files last because they were copied above
#
__ca-i_create_pkg() {
  DEST_DIR="${CDD}/distribution/${UNIQ_ID}"

  echo $UNIQ_ID > cfg/UNIQ_ID_CA-I

  # 
  # Client
  # 
  # create directories
  mkdir -p $DEST_DIR/clients/data
  mkdir -p $DEST_DIR/clients/distro
  mkdir -p $DEST_DIR/clients/docs
  mkdir -p $DEST_DIR/clients/cfg
  # copy resource files
  cp $CDD/res/libs/gen_client.sh  $DEST_DIR/clients/
  cp $CDD/res/libs/pki_funcs.sh   $DEST_DIR/clients/cfg
  cp $CDD/res/docs/README_C       $DEST_DIR/clients/README
  cp $CDD/res/docs/SERIAL_C       $DEST_DIR/clients/cfg/SERIAL
  cp "${CDD}/cfg/${ORG_URL}.cnf"  $DEST_DIR/clients/cfg/cert.cnf
  # generated files
  cp $UNIQ_ID_CAI.crt.pem     $DEST_DIR/clients/cfg/ca-i.crt.pem
  cp $UNIQ_ID_CAI.keys.pem    $DEST_DIR/clients/cfg/ca-i.keys.pem
  cp ca_cert-chain*.pem       $DEST_DIR/clients/cfg/ca_cert-chain.crts.pem
  cp cfg/UNIQ_ID_CA-I         $DEST_DIR/clients/cfg/
  cp cfg/UNIQ_ID_CA           $DEST_DIR/clients/cfg/

  # 
  # Server
  # 
  # create directories
  mkdir -p $DEST_DIR/servers/data
  mkdir -p $DEST_DIR/servers/distro
  mkdir -p $DEST_DIR/servers/docs
  mkdir -p $DEST_DIR/servers/cfg
  # copy resource files
  cp $CDD/res/libs/gen_server.sh  $DEST_DIR/servers/
  cp $CDD/res/libs/pki_funcs.sh   $DEST_DIR/servers/cfg/
  cp $CDD/res/docs/README_S       $DEST_DIR/servers/README
  cp $CDD/res/docs/SERIAL_S       $DEST_DIR/servers/cfg/SERIAL
  cp "${CDD}/cfg/${ORG_URL}.cnf"  $DEST_DIR/servers/cfg/cert.cnf
  # generated files
  cp $UNIQ_ID_CAI.crt.pem   $DEST_DIR/servers/cfg/ca-i.crt.pem
  cp $UNIQ_ID_CAI.keys.pem  $DEST_DIR/servers/cfg/ca-i.keys.pem
  cp ca_cert-chain*.pem     $DEST_DIR/servers/cfg/ca_cert-chain.crts.pem
  cp cfg/UNIQ_ID_CA-I       $DEST_DIR/servers/cfg/
  cp cfg/UNIQ_ID_CA         $DEST_DIR/servers/cfg/

  # 
  # CA-I
  #
  # create directories
  mkdir -p $DEST_DIR/ca-i/data
  mkdir -p $DEST_DIR/ca-i/docs
  mkdir -p $DEST_DIR/ca-i/distro
  # copy resource files
  cp $CDD/res/docs/README_CAI   $DEST_DIR/README
  cp $CDD/ca/*.crt.pem          $DEST_DIR/ca-i/data/
  cp $CDD/ca/*.info.txt         $DEST_DIR/ca-i/docs/
  # generated files
  mv $UNIQ_ID_CAI*.pem          $DEST_DIR/ca-i/data/
  mv $UNIQ_ID_CAI.crt.info.txt  $DEST_DIR/ca-i/docs/
  mv $UNIQ_ID_CAI.p12           $DEST_DIR/ca-i/distro
  mv ca_cert-chain*.pem         $DEST_DIR/ca-i/distro
}

# This function will generate a CA Intermediate
# 
# Requires:  CNF file, CA cert, CA key
# 
# IN: UNIQ_ID_CA
#
ca-i_gen_cert() {
  UNIQ_ID=$1
  DEST_DIR="."

  UNIQ_ID="${SERIAL}.cai.${ORG_URL}"

  echo_block "Create CA Intermediate  (${UNIQ_ID})"

  openssl genrsa -out "${DEST_DIR}/${UNIQ_ID}.keys.pem" 4096

  # Create Cert Signing Request (CSR)
  openssl req -config "cfg/ca.cnf" -new -sha256 \
              -subj "/C=OO/O=ACME/OU=ACME Intermediate/CN=${UNIQ_ID}" \
              -key "${DEST_DIR}/${UNIQ_ID}.keys.pem" -out "${DEST_DIR}/${UNIQ_ID}.csr.pem"

  # Create Certificate   (valid for ~2 years, after the entire chain of trust expires)
  # CA signs Intermediate
  openssl x509 -req -days 750 -extfile "cfg/ca.cnf" -extensions v3_ca_i \
               -CA cfg/ca.crt.pem -CAkey cfg/ca.keys.pem -set_serial ${SERIAL} \
               -in "${DEST_DIR}/${UNIQ_ID}.csr.pem" -out "${DEST_DIR}/${UNIQ_ID}.crt.pem"

  # Package the Certificate Authority Certificates for distro  (windoze needs this)
  openssl pkcs12 -export -password "pass:password" -inkey "${DEST_DIR}/${UNIQ_ID}.keys.pem" \
                 -name "CA Intermediate Mobile Provision" -certfile cfg/ca.crt.pem \
                 -in "${DEST_DIR}/${UNIQ_ID}.crt.pem" -out "${DEST_DIR}/${UNIQ_ID}.p12"

  # verify certificate (output to text file for review)
  openssl x509 -noout -text -in "${DEST_DIR}/${UNIQ_ID}.crt.pem" > "${DEST_DIR}/${UNIQ_ID}.crt.info.txt"

  # create certifiate chain
  cat cfg/ca.crt.pem "${DEST_DIR}/${UNIQ_ID}.crt.pem" > "${DEST_DIR}/ca_cert-chain_${UNIQ_ID}.crts.pem"
}

get_uniq_ids() {
  UNIQ_ID_CA=`head cfg/UNIQ_ID_CA`
  UNIQ_ID_CAI=`head cfg/UNIQ_ID_CA-I`
}

gen_client() {
  COUNT=$(($1-1))

  get_uniq_ids
  for NUM in $(seq 0 $COUNT)
  do
    get_serial
    UNIQ_ID="${SERIAL}.client.${UNIQ_ID_CAI}"
    gen_client_cert $UNIQ_ID
  done
}

#
# Generate a Client Certificate
# IN: UNIQ_ID, SERIAL
#
gen_client_cert() {
  UNIQ_ID=$1

  echo_block "Generate Client Certificates  (${UNIQ_ID})"

  openssl genrsa -out "data/${UNIQ_ID}.keys.pem" 4096

  openssl req -new -key "data/${UNIQ_ID}.keys.pem" \
              -subj "/C=OO/O=ACME/OU=ACME Standard/CN=${UNIQ_ID}" \
              -out "data/${UNIQ_ID}.csr.pem"
  # CA Intermediate signs Client
  openssl x509 -req -days 365 \
               -CA "cfg/ca-i.crt.pem" -CAkey "cfg/ca-i.keys.pem" -set_serial ${SERIAL} \
               -in "data/${UNIQ_ID}.csr.pem" -out "data/${UNIQ_ID}.crt.pem"

  # Package the Certificates
  openssl pkcs12 -export -password "pass:password" -inkey "data/${UNIQ_ID}.keys.pem" \
                 -name "Client ${UNIQ_ID} VPN Certificate" -certfile "cfg/ca_cert-chain.crts.pem" -caname "${UNIQ_ID}@acme.xyz" \
                 -in "data/${UNIQ_ID}.crt.pem" -out "distro/${UNIQ_ID}.p12"

  # verify certificate (output to text file for review)
  openssl x509 -noout -text -in "data/${UNIQ_ID}.crt.pem" > "docs/${UNIQ_ID}.info.txt"
}

gen_server() {
  COUNT=$(($1-1))

  get_uniq_ids
  for NUM in $(seq 0 $COUNT)
  do
    get_serial
    UNIQ_ID="${SERIAL}.server.${UNIQ_ID_CAI}"
    gen_server_cert $UNIQ_ID
  done
}

#
# Generate a Server Certificate
# IN: UNIQ_ID, SERIAL
#
gen_server_cert() {
  UNIQ_ID=$1

  echo_block "Generate Server Certificates  (${UNIQ_ID})"

  openssl genrsa -out "data/${UNIQ_ID}.keys.pem" 4096

  openssl req -new -config "cfg/cert.cnf" -key "data/${UNIQ_ID}.keys.pem" \
              -subj "/C=OO/O=ACME/OU=ACME Standard/CN=${UNIQ_ID}" \
              -out "data/${UNIQ_ID}.csr.pem"

  # CA Intermediate signs Server
  openssl x509 -req -days 365 -extfile "cfg/cert.cnf" -extensions v3_server \
               -CA "cfg/ca-i.crt.pem" -CAkey "cfg/ca-i.keys.pem" -set_serial ${SERIAL} \
               -in "data/${UNIQ_ID}.csr.pem" -out "data/${UNIQ_ID}.crt.pem"

  # Package the Certificates
  openssl pkcs12 -export -password "pass:password" -inkey "data/${UNIQ_ID}.keys.pem" \
                 -name "Server ${UNIQ_ID} VPN Certificate" -certfile "cfg/ca_cert-chain.crts.pem" -caname "${UNIQ_ID}@acme.xyz" \
                 -in "data/${UNIQ_ID}.crt.pem" -out "distro/${UNIQ_ID}.p12"

  # verify certificate (output to text file for review)
  openssl x509 -noout -text -in "data/${UNIQ_ID}.crt.pem" > "docs/${UNIQ_ID}.crt.info.txt" 
}

